/**
 * Demostración de refactorización, basado en el ejemplo del capítulo 1 de 
 * 'Refactoring. Improving the Design of Existing Code', de Martin Fowler.
 * 
 * Código original sin refactorizar
 */
package es.ua.dlsi.prog3.refactoring.videoclub.original;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * Cliente del videoclub. Se le identifica por su nombre.
 * 
 * Sigue las instrucciones que encontrarás en el comentario del método informe(), más abajo,
 * para refactorizar el código según las indicaciones del libro de M. Fowler.
 * 
 * En la carpeta 'test' hay un conjunto de test unitarios para comprobar que estamos refactorizando correctamente.
 * Recuerda los pasos para refactorizar:
 * - Haz una refactorización simple
 * - Pasa los test
 * 
 * Te aconsejamos que te leas al menos el primer capítulo del libro de M. Fowler.
 */
public class Cliente {

	private String nombre;
	
	/**
	 * Lista de pelis que tiene alquiladas
	 */
	private List<Alquiler> _rentals;
	
	public Cliente(String nom) {
		nombre = nom;
		_rentals = new ArrayList<>();
	}

	public String getNombre() { return nombre; }
	
	/**
	 * Añade un nuevo alquiler al cliente
	 * @param a el alquiler
	 */
	public void addAlquiler(Alquiler a) {
		_rentals.add(a);
	}
	
	/**
	 * Refactorización:
	 * 0. Copiar el paquete entero, es.ua.dlsi.prog3.refactoring.videoclub.original, en otro llamado
	 *    es.ua.dlsi.prog3.refactoring.videoclub.original.copy. Refactorizar este último.
	 *
	 * 1. Encapsular el switch -> método double calcularCargo(Alquiler)
	 *
	 * 2. El cargo normal depende del tipo de peli:
	 *    2.1 : Crear subclases de Pelicula: PeliculaInfantil, PeliculaNormal, PeliculaEstreno
	 *    2.2 : Introducir el método abstracto Pelicula.getPrecio() y sus implementaciones en las subclases
	 *
	 * 3. El máximo de días de alquiler también depende del tipo de película. Crear método abstracto
	 *    Pelicula.getMaxDiasAlquiler() y su implementación en las subclases. Al hacer esto,
	 *    el código del switch para peliculas normales e infantiles queda idéntico.
	 *
	 * 4. El cálculo del cargo por alquiler depende de dos cosas: del tipo de película y del número de días de alquiler.
	 *    Crear el método 'double Pelicula.calculaCargoAlquiler(Alquiler a)' que refactorice el código
	 *    de los case del switch. La implementación por defecto en película puede ser la de peli normal e infantil,
	 *    sobreescribiendolo en PeliculaEstreno.
	 *
	 * 5. Ahora ya estamos en disposición de eliminar el switch en Cliente.calcularCargo(Alquiler) por
	 *      return each.getPelicula().calculaCargoAlquiler(each);
	 * 5b. Un momento... Cliente.calcularCargo(Alquiler) no necesita nada de Cliente...
	 *      muévelo a Alquiler.calcularCargo()
	 *
	 * 6. A partir de ahora, podemos añadir nuevos tipos de peli, cambiar el modo en que se calcula el cargo, introducir
	 *    cambios en los precios... sin tocar para nada el código de Cliente.informe(). 
	 * 6b. Podemos eliminar los codigos de precio (INFANTIL, ESTRENO...) y los parámetros y métodos asociados de la jerarquía
	 *    de películas. Ya no son necesarios.
	 *
	 * 7. Añade un nuevo tipo de pelicula, por ejemplo, 'PeliSerieB' al sistema. Recuerda cambiar de sombrero ;-)   
	 *
	 * 8. Descomentar el método es.ua.dlsi.prog3.refactoring.videoclub.Videoclub.runRefactored() y su llamada desde main()
	 *    para comprobar que, tras la refactorización, el comportamiento de Cliente.informe() es el mismo
	 *    
	 * 9. PREGUNTA: ¿Que hemos ganado con la refactorización, si el código hace exactamente lo mismo?
	 * 
	 * 10. PREGUNTA: ¿Qué más refactorizaciones podríamos aplicar a Cliente.informe()? 
	 *     PISTA: Piensa en que tenemos que implementar una versión del método que imprima el informe el formato HTML
	 *     
	 * 11. EJERCICIO: Refactoriza el código ue calcula los puntos por alquiler frecuente (frequentRenterPoints)
	 * 
	 * @return el informe en formato texto plano
	 */
	public String informe() {
		double totalAmount = 0;
		int frequentRenterPoints = 0;
		Iterator<Alquiler> rentals = _rentals.iterator();
		String result = "Rental Record for " + getNombre() + "\n";

		while (rentals.hasNext()) {
			double thisAmount = 0;
			Alquiler each = rentals.next();

			//determine amounts for each line
			switch (each.getPelicula().getCodigoPrecio()) {
			case Pelicula.NORMAL:
				thisAmount += 2;
				if (each.getDiasAlquiler() > 2) 							
					thisAmount += (each.getDiasAlquiler() - 2) * 1.5;
				break;
			case Pelicula.ESTRENO:
				thisAmount += each.getDiasAlquiler() * 3;
				break;
			case Pelicula.INFANTIL:
				thisAmount += 1.5;
				if (each.getDiasAlquiler() > 3)
					thisAmount += (each.getDiasAlquiler() - 3) * 1.5;
				break;
			}

			// add frequent renter points
			frequentRenterPoints++;
			// add bonus for a two day new release rental
			if ((each.getPelicula().getCodigoPrecio() == Pelicula.ESTRENO) 
					&& each.getDiasAlquiler() > 1) 
				frequentRenterPoints++;

			// show figures for this rental
			result += "\t" + each.getPelicula().getTitulo() 
					+ "\t" + String.valueOf(thisAmount) + "\n";

			totalAmount += thisAmount;
		} // END WHILE

		// add footer lines
		result += "Amount owed is " + String.valueOf(totalAmount) + "\n";
		result += "You earned " + String.valueOf(frequentRenterPoints) 
		+ " frequent renter points";

		return result;
	}


	/**
	 * Genera un informe en HTML de lo que debe el usuario
	 * @return el informe
	 */
	public String informeHTML() {
		double totalAmount = 0;
		int frequentRenterPoints = 0;
		Iterator<Alquiler> rentals = _rentals.iterator();
		String result = "<html></html><body>\n"+ 
				"<p>Rental Record for " + getNombre() + "</p>\n";
		result += "<ul>\n";
	
		while (rentals.hasNext()) {
			double thisAmount = 0;
			Alquiler each = rentals.next();
	
			//determine amounts for each line
			switch (each.getPelicula().getCodigoPrecio()) {
			case Pelicula.NORMAL:
				thisAmount += 2;
				if (each.getDiasAlquiler() > 2) 							
					thisAmount += (each.getDiasAlquiler() - 2) * 1.5;
				break;
			case Pelicula.ESTRENO:
				thisAmount += each.getDiasAlquiler() * 3;
				break;
			case Pelicula.INFANTIL:
				thisAmount += 1.5;
				if (each.getDiasAlquiler() > 3)
					thisAmount += (each.getDiasAlquiler() - 3) * 1.5;
				break;
			}
	
			// add frequent renter points
			frequentRenterPoints++;
			// add bonus for a two day new release rental
			if ((each.getPelicula().getCodigoPrecio() == Pelicula.ESTRENO) 
					&& each.getDiasAlquiler() > 1) 
				frequentRenterPoints++;
	
			// show figures for this rental
			result += "<li>" + each.getPelicula().getTitulo() 
					+ " " + String.valueOf(thisAmount) + "</li>\n";
	
			totalAmount += thisAmount;
		} // END WHILE
	
		// add footer lines
		result += "</ul>";
		result += "<p>Amount owed is " + String.valueOf(totalAmount) + "</p>\n";
		result += "<p>You earned " + String.valueOf(frequentRenterPoints) 
		+ " frequent renter points</p>\n</body>\n";
	
		return result;
	}
	
}
